;;; <genetische-Algorithmen.scm>
(load "breakpoint.scm") ; bei neuer Version writeln -> ausgabe

; Mutationswahrscheinlichkeit und Selektionswahrscheinlichkeit
; sind positive ganze Zahlen zwischen 0 und 100 ( also % )

;;; ===== Entwurfsentscheidungen ====================================
; Auch wenn der Name 'genetischer-Algorithmus' anderes andeutet, es
; gibt nicht den genetischen Algorithmus, dies ist >>>einer<<< ! 
; Zu den einzelnen Entwurfsentscheidungen siehe auch bei den
; einzelnen Teilprozessen.
; In der Funktion selbst steckt z.B. die Entwurfsentscheidung, den
; Abbruch allein von der Zahl der Schritte abhaengig zu machen. Man
; kann auch andere Abbruchkriterien waehlen.

; Eine weitere wichtige Entwurfsentscheidung betrifft die Bewertung
; der Individuen in den Populationen. Wenn auch der eigentliche Sinn
; der GA fuer die Informatik in der Tatsache steckt, dass die Bewer-
; tungen erheblich 'billiger' sind als das Generieren von Loesungen,
; steckt in der Verwaltung dieser Werte ein abzuwaegender Verwaltungs-
; aufwand. Ich habe mich hier daher so entschieden, waehrend der
; Bearbeitung die Bewertungen an die Individuen zu koppeln.
; Jedes von ihnen stellt waehrend der Bearbeitung daher ein Paar 
; (Individuum Bewertung) dar!
; Das setzt voraus, dass beim jedem Erzeugen eines neuen Individuums
; sofort eine Bewertung durchgefuehrt wird.

;;; ===== genetischer-Algorithmus ===================================
(define
  (genetischer-Algorithmus 
   Populations-Groesse
   Anzahl-Schritte
   Mutationswahrscheinlichkeit
   crossing-over-Wahrscheinlichkeit)
  (let
      loop
    ((Population 
      (Bewertung (erzeuge-Population Populations-Groesse)))
     (zaehler Anzahl-Schritte))
    (Darstellung (bestes-Individuum Population)) 
    (ausgabe "Population : " Population (length Population))
    (cond
      ((zero? zaehler)
        (bestes-Individuum Population))
      (else
       (loop
        (Selektion
         (Kreuzen
          (Mutation Population Mutationswahrscheinlichkeit)
          crossing-over-Wahrscheinlichkeit)
         Populations-Groesse)
        (sub1 zaehler))))))


;;; ===== GA-Aufruf =================================================
; Eine Aufrufhuelle fuer 'genetischer-Algorithmus', bei der die
; Parameter zur Laufzeit eingegeben werden koennen
(define
  (GA-Aufruf)
  (let
      ((Populations-Groesse
       (begin
         (display "Gib die Zahl der Individuen an [> 0] : ")
         (read)))
       (Anzahl-Schritte
       (begin
         (display "Wieviel Schritte ?             [> 0] : ")
         (read)))
       (Mutationswahrscheinlichkeit
       (begin
         (display "Mutationswahrscheinlichkeit ?  [00 .. 99] : ")
         (read)))
       (crossing-over-Wahrscheinlichkeit
        (begin
          (display 
           "Wahrscheinlichkeit fuer crossing-over ? [00 .. 99] : ")
          (read))))
    (genetischer-Algorithmus 
     Populations-Groesse
     Anzahl-Schritte
     Mutationswahrscheinlichkeit
     crossing-over-Wahrscheinlichkeit)))

;;; ===== erzeuge-Population ========================================
; Erzeugt die Grundpopulation fuer den genetischen Prozess.
; Parameter sind: Anzahl der Individuen,
; Wert: Eine Liste von Individuen
(define
  (erzeuge-Population Anzahl)
  (let
      loop
    ((Population ())
     (n Anzahl))
    (cond
      ((zero? n)
       Population)
      (else
       (loop
        (cons
         (erzeuge-Individuum)
         Population)
        (sub1 n))))))

;;; ===== erzeuge-Individuum ========================================
; Erzeugt ein Individuum der Population.
; Dieser eigentliche Erzeugungsprozess ist problemabhaengig.
(define
  (erzeuge-Individuum)
  (random 100))
;;; ===== aktion? ===================================================
; Hilfsfunktion, die entscheidet ob eine Aktion stattfindet oder
; nicht. Die Entscheidung haengt von der uebergebenen ganzzahligen
; Wahrscheinlichkeit p ab.
(define
  (aktion? p)
  (< (random 100) p))

;;; ===== waehle-Positionen =========================================
; Waehlt zwei verschiedene Zahlen aus {0 .. n-1} und gib sie als Liste
; zurueck. Wird als Hilfsfunktion benoetigt.
; Die Zahlen sind stets geordnet: (klein gross)
; Unzulaessige Parameter fuer n loesen einen Fehler aus!
(define
  (waehle-Positionen n)
  (cond
    ((not (number? n)) (error "keine Zahl !" n))
    ((not (integer? n)) (error "keine ganze Zahl !" n))
    ((< n 2) (error "keine Auswahl moeglich !" n))
    (else
     (let
         loop
       ((i (random n))
        (j (random n)))
       (cond
         ((and (= i j) (positive? i)) (list (sub1 i) j))
         ((= i j) (list i (add1 j)))
         ((> i j) (list j i))
         (else
          (list i j)))))))

;;; ===== Mutation ==================================================
; Steuert die Mutation fuer die gesamte Population.
; Die bei der Mutation neu erzeugten Individuen werden der Population
; hinzugefuegt, dadurch aendert sich die Gesamtzahl. Deren Anpassung
; erfolgt bei der Selektion.
; Parameter sind: Population, Mutationswahrscheinlichkeit
; Wert: mutierte Population.
(define
  (Mutation Population Mutationswahrscheinlichkeit)
  (let
      loop
    ((Population Population)
     (akku ()))
    (cond
      ((null? Population)
       (reverse akku))
      ((aktion? Mutationswahrscheinlichkeit)
       (loop
        (rest Population)
        (cons
         (bewerte (mutiere (first (first Population))))
         (cons
          (first Population)
          akku))))
      (else
       (loop
        (rest Population)
        (cons
         (first Population)
         akku))))))

;;; ===== mutiere ===================================================
; Mutiert ein einzelnes Individuum.
; Hier muss die Bewertung bei den Individuen fehlen !
; Dieser eigentliche Mutationsprozess ist problemabhaengig.
; Individuum -> Individuum
(define
  (mutiere Individuum)
  (quotient (+ (random 100) Individuum) 2))

;;; ===== Kreuzen ===================================================
; Steuert das Kreuzen fuer die gesamte Population.
; Da beim Kreuzen neue Individuen erzeugt und der Population hinzu-
; gefuegt werden, aendert sich die Gesamtzahl. Deren Anpassung erfolgt
; bei der Selektion.
; Die zu kreuzenden Individuen werden nach dem Zufallsprinzip ausge-
; waehlt.
; Parameter sind: Population, crossing-over-Wahrscheinlichkeit
; Wert: gekreuzte Population.
(define
  (Kreuzen Population crossing-over-Wahrscheinlichkeit)
  (let
      ((n (length Population)))
    (let
        loop
      ((Auswahlen (waehle-Positionen n))
       (i (length Population))                ; Zaehler
       (akku ()))
      (cond
        ((zero? i)
         (append akku Population))
        ((aktion? crossing-over-Wahrscheinlichkeit)
         (let
             ((neue 
               (kreuze
                (first (list-ref Population (first Auswahlen)))
                (first (list-ref Population (second Auswahlen))))))
           (loop
            (waehle-Positionen n)
            (sub1 i)
            (cons
             (bewerte (first neue))
             (cons
              (bewerte (second neue))
              akku)))))
        (else
         (loop
          (waehle-Positionen n)
          (sub1 i)
          akku))))))

;;; ===== kreuze ====================================================
; Kreuzt zwei einzelne Individuen.
; Hier muss die Bewertung bei den Individuen fehlen !
; Dieser eigentliche Kreuzungsprozess ist problemabhaengig.
; Individuum x Individuum -> (list Individuum Individuum)
(define
  (kreuze Individuum-1 Individuum-2)
  (list 
    (quotient (+ Individuum-1 Individuum-1 Individuum-1 Individuum-2) 4)
    (quotient (+ Individuum-1 Individuum-2 Individuum-2 Individuum-2) 4)))

;;; ===== Selektion =================================================
; Es gibt viele veschiedene Selektionsmechanismen. Beispielsweise
; kann die Auswahl proportional zur Bewertungsfunktion erfolgen.
; Hier werden die vorhandenen Individuen einfach zufaellig ausgewaehlt
; und einer paarweisen Auswahl unterzogen: Selektiert wird aus der
; Menge der Individuen durch direkten Vergleich von je zwei
; Individuen. Der Parameter Gesamtzahl gewaehrleistet, dass wieder
; die urspruengliche Populationsgroesse hergestellt wird.
; Dazu wird mit der Selektion abgebrochen, wenn eine genuegend grosse
; Zahl ausgewaehlt wurde.
(define
  (Selektion Population Gesamtzahl)
  (let
      loop
    ((Auswahlen (waehle-Positionen (length Population)))
     (n (length Population))
     (akku ()))
    (cond
      ((= Gesamtzahl (length akku))
       akku)
      (else
       (loop
        (waehle-Positionen n)
        n
        (cons
         (selektiere
          (list-ref Population (first Auswahlen))
          (list-ref Population (second Auswahlen)))
         akku))))))

;;; ===== selektiere ================================================
; Selektiert von zwei Individuen das mit der hoeheren Bewertungszahl.
; Obwohl hier die Bewertungen der Individuen maximal zweimal bestimmt
; werden, kann es sich lohnen, diese Werte zunaechst in einer neuen
; Liste oder in der Populationsliste mit abzulegen.
; So kann z.B. auch in der Ausgabe darauf zugegriffen werden.
; siehe-> Bewertung
(define
  (selektiere Individuum-1 Individuum-2)
  (if
   (besser? Individuum-1 Individuum-2)
   Individuum-1
   Individuum-2))

;;; ===== besser? ===================================================
; Das Praedikat 'besser?' ist das Vergleichspraedikat zur Bewertung und
; ist daher vom speziellen Problem abhaengig. Hier wird einfach die
; hoehere Bewertungszahl genommen.
(define
  (besser? Individuum-1 Individuum-2)
   (> (second Individuum-1) (second Individuum-2)))

;;; ===== Bewertung =================================================
; Fuehrt die Bewertung der gesamten Population durch. Dazu wird die 
; Populationsliste in eine Assoziationsliste mit Paaren 
; (Individuum Bewertung) verwandelt.
(define
  (Bewertung Population)
  (let
      loop
    ((Population Population)
     (akku ()))
    (cond
      ((null? Population)
       (reverse akku))
      (else
       (loop
        (rest Population)
        (cons
         (bewerte (first Population))
         akku))))))

;;; ===== bewerte ===================================================
; Bewertet ein Individuum mit einer Bewertungsziffer.
; Liefert eine Liste aus Individuum und Bewertungszahl zurueck
; Die Bewertungsfunktion haengt vom speziellen Problem ab.
(define
  (bewerte Individuum)
  (list Individuum (* 10 Individuum)))

;;; ===== bestes-Individuum =========================================
; Fuer die grafische Darstellung, aber auch fuer das Ergebnis des GA
; ist es sinnvoll, das 'beste' Individuum bestimmen zu koennen.
(define
  (bestes-Individuum Population)
  (let
      loop
    ((Population Population)
     (bestes (first Population)))
    (cond
      ((null? Population) bestes)
      ((besser? (first Population) bestes)
       (loop (rest Population) (first Population)))
      (else
       (loop (rest Population) bestes)))))

;;; ===== Darstellung ===============================================
; Die Darstellung der 'besten' Loesung ist problemabhaengig.
(define
  (Darstellung Individuum)
  (display Individuum)
  (newline))

;;; ===== Zufallsgenerator-setzen ===================================
; Setzt den Zufallsgenerator je nach Parameter mit einer "zufaelligen"
; Zahl oder der Zahl 123456789.
(define 
  (Zufallsgenerator-setzen Zeit?)
  (random-seed 
   (if
    Zeit?
    (remainder 
     (abs (current-milliseconds))
     1000000000)
    123456789)))


;;; ===== Aufruf ====================================================
(Zufallsgenerator-setzen #t)
(genetischer-Algorithmus 5 10 20 30)
